/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/******************************************************************************
 * defines functions 
 * - to update the contents of the history tracers editor 
 * - to keep the internal command statistics
 *
 * notice:
 * The function 'updateStatistics' needs as
 * parameters the number of lines which are inserted or deleted and the
 * line where the changes start. That means 
 * - new lines are inserted behind the given line; new lines are
 * always added to the previous command, so no new commands can be
 * inserted by editing the commands editor
 * -  deletion starts at the given line and deletes the lines from
 * 'line' to 'line - numberOfLines'; i.e. deletion is executed backwards
 * from the given position. 
 * 
 * The command statistics is run through from the last command to the first
 * one searching for the commands affected by the changes. All commands
 * beginning completely after the changes have to be increased by the number
 * of lines changed. All the commands with a smaller index than the affected
 * ones need not be changed. A command can be affected by a change in 4 ways:
 * 1) the lower border of a command is crossed by deletions
 * (the current command is marked as open; the rest of this command is
 * appended to the previous command in case 2); the current
 * command entry is marked as deleted)
 * 2) the upper border of a command is crossed by deletions
 * (the upper border is set to the line up to which the command is
 * deleted or to the end of an open command minus the deleted range if
 * case 1) occurred)
 * 3) a complete command is deleted
 *(the corresponding entry in the statistics is marked as deleted; in
 * the end all marked entries are deleted and the following entries are
 * moved up)
 * 4) changes are local to a command
 * (upper border has to be adapted)
 *
 * In cases 1) and 4) the run through the command statistics can be
 * stopped, because the earlier commands need not be changed.
 *****************************************************************************/

#include <stdio.h>
#include "HT.h"
#include "../exEdit/display.h"

BufInfo *bufinfoptr();
int ETnumlines();

void saveComIndex(); 
static insertComNum(); 
static delComInSeq();
static delHistComm(); 
static updateNumEdit();

static int COMPTR = 0;
static char msg[200];

/* nextCommand ****************************************************************
 * handle string up to next SOC signal if one exists and return rest of string
 *****************************************************************************/

char *nextCommand(htracer, socPtr)
     Editor htracer;
     char *socPtr;
{
  char *eoc,              /* points to character following EOC */
       *curcom;           /* holds string between two signals */
  char *index();

  eoc = ++socPtr;                       
  if (*socPtr == EOS) 
    return(socPtr);
	
  if ((socPtr = index(socPtr, SOC)) == NULL) 
    curcom = strCopy(eoc);
  else {	                                               /* SOC signal */
    curcom = makeStr(eoc, socPtr, strlen(eoc) - strlen(socPtr) + 1);
    socPtr++;                                  /* position behind SOC signal */
  }

  /* commands for history tracer */
  edit(bufinfoptr(htracer), curcom, strlen(curcom));                         
  free(curcom);

  return(socPtr);
}

/* addCommToHist *************************************************************
 * append complete command to history tracer; if last line of text is visible
 * show new command
 ****************************************************************************/

addCommToHist(hw, clist, comm)
     historyWindow *hw;
     commandStatistics *clist;
     char *comm;                     /* command to add to history */
{
  int noLines;                       /* number of lines in command */
  register BufInfo *bf = bufinfoptr(hw->comEdit);

  if ((noLines = cntNewLines(comm)) == 0)
    return;

  autodelHist(hw, clist, noLines);

  saveComIndex(hw->comEdit, clist, ETnumlines(bf) - 1, START);
  editorAppend(hw->comEdit, comm);     /* cursor is positioned to the end of 
					  the last line of the new command; 
					  right/left scrolling is done if 
					  necessary */

  COMPTR++;
  insertComNum(hw->numEdit, COMPTR /*clist->comCount + 1*/, noLines);
  saveComIndex(hw->comEdit, clist, ETnumlines(bf) - 1, END);

  newline(hw->comEdit);
  ESsetcur(bf);

  setHistScrollbar(hw);
}

/* autodelHist ***************************************************************
 * delete the first commands from history tracer in order to append new 
 * commands;
 ****************************************************************************/

void autodelHist(hw, cl, lines)
     historyWindow *hw;
     commandStatistics *cl;
     int lines;
{
  register BufInfo *bf = bufinfoptr(hw->comEdit);
  static Boolean htFull = False;     /* indicates if history tracer is full */
  
  if (ETnumlines(bf) + lines >= MAXHIST) {

    if (!htFull) {           /* inform user that first commands are deleted */
      sprintf(msg, "History Tracer is full;\n the first commands will be\n deleted successively! Do you want\n to save current history\n on file first?");
         if (exNotifyUser(hw->histShell, -1, -1, msg) == ANSWERYES) 
	saveHistory((RequestParm *) NULL);      
      htFull = True;
    }

    delHistComm(hw, cl, MAX(MAXHIST/10 + 1, lines + 1));
  }
}

/* delHistComm ***************************************************************
 * delete the first commands from history tracer in order to append new 
 * commands;
 ****************************************************************************/

static delHistComm(hw, clist, lines)
     historyWindow *hw;
     commandStatistics *clist;
     int lines;                      /* number of lines to be deleted */
{
  register i = 0, j, dellines = 0;
  register BufInfo *bf = bufinfoptr(hw->comEdit);
  register BufInfo *bfe = bufinfoptr(hw->numEdit); 

  while (dellines < lines) {                  /* delete first commands */
    dellines += clist->comSeq[i][1] - clist->comSeq[i][0] + 1;
    i++;
  }

  autodeleteLines(bf, dellines);
  autodeleteLines(bfe, dellines);
  
  for (j = 0; j < i; j++) {      /* remove deleted commands from statistics */
    clist->comSeq[j][0] = -1;
    clist->comSeq[j][1] = -1;
  }

  delComInSeq(clist);            /* rearrange statistics */

  for (j = 0; j < clist->comCount; j++) {        /* adjust remaining commands 
						    in statistics */
    clist->comSeq[j][0] -= dellines;
    clist->comSeq[j][1] -= dellines;
  }    

  adjustHistPtr(-i);
}

/* saveComIndex **************************************************************
 * store index of text line a command is starting/ending with
 ****************************************************************************/

void saveComIndex(editor, cs, line, start)
     Editor editor;
     commandStatistics *cs;      /* command statistics */
     int line;                   /* line number to be stored */
     Boolean start;              /* if TRUE, line marks beginning of command */
{
  if (start) 
    cs->comSeq[cs->comCount][0] = line;
  else {
    cs->comSeq[cs->comCount][1] = line;
    cs->comCount++;
    deleteComplCom(cs);
  }
}

/* insertComNum **************************************************************
 * insert command number and line marks into number editor
 ****************************************************************************/

static insertComNum(nEdit, cn, comLines)
     Editor nEdit;
     int cn;                           /* number to be inserted */
     int comLines;                     /* number of lines of new command */
{ 
  register i;
  char ident[8];

  if (cn < 10)
    sprintf(ident, " %d:", cn);
  else
    sprintf(ident, "%d:", cn);

  editorAppend(nEdit, ident);
  newline(nEdit);

  for (i = 1; i < comLines; i++) {
    editorAppend(nEdit, LINEMARK);
    newline(nEdit);
  }

  putCursor(nEdit);
}

/* updateStatistics **********************************************************
 * update number of command and indices marking the beginning and the end of a 
 * command (added lines are always appended to an existing command; i.e. no new
 * commands can be inserted from outside); 
 * changes start at "line" and affect the following/previous lines
 * - delete lines: range 'line' to '(line - abs(diff))' is deleted,
 * - insert lines: new lines are inserted behind 'line'.
 *****************************************************************************/

updateStatistics(ht, cs, line, diff)
     historyWindow *ht;
     commandStatistics *cs;
     int line,                               /* line where changes started */
         diff;                               /* number of lines changed */
{
  register i, 
           openCom = -1;                     /* index of open command */
  Boolean change = False;

  for (i = cs->comCount - 1; i >= 0; i--) {
    if (diff > 0 && cs->comSeq[i][0] <= line && cs->comSeq[i][1] >= line) {  
	                                         /* add lines into a command */
      updateNumEdit(ht, cs->comSeq[i][0], diff);
      cs->comSeq[i][1] += diff;
      break;
    }

    else if (diff < 0 && cs->comSeq[i][0] <= line - abs(diff)
	     && cs->comSeq[i][1] >= line) {   
                                               /* delete lines in a command */
      updateNumEdit(ht, cs->comSeq[i][1], diff);
      cs->comSeq[i][1] += diff;
      break;
    }

    else if (diff < 0 && cs->comSeq[i][0] <= line 
	     && cs->comSeq[i][0] > line - abs(diff) && cs->comSeq[i][1] > line)
      {                     /* lower border of command is crossed by changes */

	updateNumEdit(ht, line, cs->comSeq[i][0] - line - 1);

	openCom = i;           /* indicates that the last part of this command 
				  has to be joint with the previous command */
      }

    else if (diff < 0 && cs->comSeq[i][1] <= line 
	     && cs->comSeq[i][1] >= line - abs(diff) 
	     && cs->comSeq[i][0] <= line - abs(diff)) {
                           /* upper border of command is crossed by changes */

      updateNumEdit(ht, line, line - abs(diff) - cs->comSeq[i][1]);

      if (openCom > 0) {          /* if open command exists join it 
				     with current command */
	/*updateNumEdit(ht, line, abs(diff));*/
	cs->comSeq[i][1] = cs->comSeq[openCom][1] - abs(diff);

	cs->comSeq[openCom][0] = -1;          /* and mark old one as deleted */
	cs->comSeq[openCom][1] = -1;
	change = True;
      }
      break;
    }
      
    else if (diff < 0 && cs->comSeq[i][1] <= line
	     && cs->comSeq[i][0] > line - abs(diff)) {
                              /* changes delete more than a complete command */
      updateNumEdit(ht, cs->comSeq[i][1], 
		    cs->comSeq[i][0] - cs->comSeq[i][1] - 1);

      cs->comSeq[i][0] = -1;
      cs->comSeq[i][1] = -1;
      change = True;
    }

    else if (cs->comSeq[i][0] > line) { 
                             /* adapt complete command stored behind changes */
      cs->comSeq[i][0] += diff;
      cs->comSeq[i][1] += diff;
    }
  }

  if (change)
    delComInSeq(cs);

  setHistScrollbar(ht);
}

/* updateNumEdit: ************************************************************
 * update of number editor: 
 * - insert linemark if new lines are inserted
 * - delete lines in number editor which correspond to lines in history editor
 ****************************************************************************/

static updateNumEdit(hs, row, range)
     historyWindow *hs;
     int row,                      /* row in editor where changes occur */
         range;                    /* number of lines changed */
{
  register i;
  register BufInfo *bf = bufinfoptr(hs->numEdit); 
  register BufInfo *bfe = bufinfoptr(hs->comEdit);

  if (range > 0) {

    if (bf->topline < bfe->topline)
      ESscroll(bf, range, DIR_UP); 

    bf->curline = row;
 
    for (i = 0; i < range; i++) {
      openlnd(hs->numEdit);                 /* endlin(hs->numEdit);
					       newline(hs->numEdit);*/
      editorInsertAt(hs->numEdit, LINEMARK, bf->curline, 0);
    } 
  }
  else if (range < 0) {                                   /* delete forward */
      
    if (bf->topline > bfe->topline)
      ESscroll(bf, abs(range), DIR_DOWN);

    bf->curline = row;

    for (i = range; i < 0; i++) {
      ETdelline(bf, row);      
      if (row - bf->topline < bf->lines - 1)       /* deleted line visible? */
	ESdelrow(bf, row - bf->topline);
      row--;
    }
    bf->currow = bfe->currow;
  }

  ESsetcur(bf);
}
 
/* delComInSeq ***************************************************************
 * run through command sequence array and delete command reference
 *****************************************************************************/

static delComInSeq(cStat)
     commandStatistics *cStat;
{
  register i = 0, j;
  
  while (i < cStat->comCount) {
    if (cStat->comSeq[i][0] == -1 && cStat->comSeq[i][1] == -1) {
                                                     /* empty command */
      for (j = i; j < cStat->comCount; j++) {
	cStat->comSeq[j][0] = cStat->comSeq[j + 1][0];
	cStat->comSeq[j][1] = cStat->comSeq[j + 1][1];
	cStat->comSeq[j + 1][0] = -1;
	cStat->comSeq[j + 1][1] = -1;
      }
      cStat->comCount--;
    }
    else
      i++;
  }
}

/* deleteStatistics ***********************************************************
 * delete command references
 *****************************************************************************/

deleteStatistics(cStat)
     commandStatistics *cStat;
{
  register i;
  
  for (i = 0; i < cStat->comCount; i++) {
    cStat->comSeq[i][0] = -1;
    cStat->comSeq[i][1] = -1;
  }

  cStat->comCount = 0;
  deleteComplCom(cStat);
}

/* compoundCom ***************************************************************
 * complete command for history tracer
 *****************************************************************************/

compoundCom(clist, buffer)
     commandStatistics *clist;
     char *buffer;
{
  char *realloc();

  if (! clist->complCom)
    clist->complCom = Malloc(strlen(buffer) + 1);
  else 
    clist->complCom = realloc(clist->complCom, 
			      strlen(clist->complCom) + strlen(buffer) + 1); 

  strcat(clist->complCom, buffer);
}

/* deleteComplCom ************************************************************
 * delete composed command
 ****************************************************************************/

deleteComplCom(cs)
     commandStatistics *cs;
{
  if (cs && cs->complCom) {
    free(cs->complCom);
    cs->complCom = NULL;
  }
}

/* cntNewLines  ***************************************************************
 * count number of '\n' in buffer and calculate numbers of text lines 
 *****************************************************************************/
 
cntNewLines(buffer)
     char *buffer;
{
  int count = 0;
  char *index();

  if (strlen(buffer) == 0)
    return(count);

  count++;
  if (*buffer == '\n')            /* first line of text */
    count++;

  while ((buffer = index(buffer + 1, '\n')) != NULL)
    count++;

  return(count);
}

/* setHistScrollbar **********************************************************
 * procedure to set scrollbar relative to text size 
 ****************************************************************************/

void setHistScrollbar(hwindow)
     historyWindow *hwindow;
{
  register  BufInfo *bf = bufinfoptr(hwindow->comEdit);
  int maxline, value, size, incr, page_incr;

  maxline = ETnumlines(bf);  
  size  = MIN((int) ((float) bf->height * (float) bf->lines / (float) maxline),
	      bf->height);
  value = MIN(bf->height - size, 
	      (int) ((float) (bf->height * bf->topline) / (float) maxline));
  incr = (int) ((float) bf->height / (float) maxline);
  page_incr = MIN(bf->height, MIN(bf->lines, ETnumlines(bf)) * incr);

  XmScrollBarSetValues(hwindow->scrollbar, value, size, incr, page_incr, True);
}

/* getComIndex ****************************************************************
 * find index of entry in the command statistics which includes specified line
 ****************************************************************************/

getComIndex(curLine, cs)
     int curLine;                       /* line of a command */
     commandStatistics *cs;             /* command statistics */
{
  register i;

  for (i = cs->comCount - 1; i >= 0; i--)
    if (cs->comSeq[i][0] <= curLine && cs->comSeq[i][1] >= curLine) 
	                                     /* curLine in current command */
      return(i);

  return(-1);
}





